import os
import re
import shutil
import tempfile
import tomllib as toml
import uuid
from dataclasses import dataclass, field
from logging import getLogger
from pathlib import Path
from typing import Optional, Sequence, Callable

from pi_haiku.models import (
    BuildSystem,
    BuildSystemError,
    PackageMatch,
    PathType,
    PyPackage,
)
from pi_haiku.utils.utils import (
    find_duplicates,
    run_bash_command,
    special_substitutions,
)

from pi_haiku.utils.environment_utils import EnvHelper
log = getLogger(__name__)
EXCLUDE_DIRS: list[PathType] = ["__pycache__", "dist", "docker_staging"]
lsentinel = [uuid.uuid4().hex]


@dataclass
class PyProjectModifier:
    src: PathType
    pyproj: PyPackage = field(init=False)
    package_dir: Optional[PathType] = None
    packages: dict[str, PyPackage] = field(default_factory=dict)
    exclude_dirs: list[PathType] = field(default_factory=lambda: EXCLUDE_DIRS)
    exclude_hidden: bool = True

    def __post_init__(self) -> None:
        self.src = Path(self.src).expanduser().resolve()
        if not self.src.exists():
            raise FileNotFoundError(f"Could not find the file {self.src}")
        with open(self.src, "rb") as fp:
            toml.load(fp)
        self.pyproj = PyPackage.from_path(self.src)
        if self.pyproj is None:
            raise ValueError(f"Could not load the pyproject.toml file: {self.src}")
        if self.package_dir and not self.packages:
            self.package_dir = Path(self.package_dir).expanduser().resolve()
            tomls = self.find_pyproject_tomls(self.package_dir, exclude_dirs=self.exclude_dirs)

            self.packages = self._tomls_to_packages(tomls)

    @staticmethod
    def _tomls_to_packages(tomls: Sequence[PathType]) -> dict[str, PyPackage]:
        pkgs = {}
        found = []
        for f in tomls:
            try:
                pkg = PyPackage.from_path(f)
                pkgs[pkg.name] = pkg
                found.append(pkg)
            except BuildSystemError as e:
                pass
            except Exception as e:
                log.error(f"File {f} had the following Error occurred: {e}")
                raise
        duplicates = find_duplicates(found)
        if duplicates:
            raise ValueError(f"Found duplicate packages: {duplicates}")
        return pkgs

    @staticmethod
    def find_pyproject_tomls(
        base_directory: PathType,
        exclude_dirs: Optional[Sequence[PathType]] = lsentinel,
        exclude_hidden: bool = True,
        file_match: str = "pyproject.toml",
    ) -> list[Path]:
        base_path = Path(base_directory).expanduser().resolve()
        pyproject_files = []

        if exclude_dirs is lsentinel:
            exclude_dirs = EXCLUDE_DIRS
        elif exclude_dirs is None:
            exclude_dirs = []
        set_exclude_dirs = set(exclude_dirs)

        for root, dirs, files in os.walk(base_path):
            # Check if the current directory should be excluded
            if any(excluded in Path(root).parts for excluded in set_exclude_dirs):
                continue

            # Filter out directories to skip
            dirs[:] = [
                d
                for d in dirs
                if (d not in set_exclude_dirs) and (not exclude_hidden or not d.startswith("."))
            ]

            if file_match in files:
                pyproject_files.append(Path(root) / file_match)

        return pyproject_files

    @staticmethod
    def find_pyprojects(
        base_directory: PathType,
        exclude_dirs: Optional[list[str]] = lsentinel,
        build_system: BuildSystem = BuildSystem.POETRY,
    ) -> dict[str, PyPackage]:
        tomls = PyProjectModifier.find_pyproject_tomls(base_directory, exclude_dirs)
        return PyProjectModifier._tomls_to_packages(tomls)

    def convert_to_remote(
        self,
        match_patterns: Optional[Sequence[PackageMatch]] = None,
        packages: Optional[Sequence[PyPackage]] = None,
        dest_file: Optional[str] = None,
        in_place: bool = False,
        use_toml_sort: bool = True,
        update: bool = False,
        backup_dir: Optional[PathType] = None,
        should_change_module: Callable[[str], bool] = lambda _: True,
    ) -> list[tuple[str, str]]:
        """
        Convert package dependencies to remote (published) versions.

        Args:
            match_patterns (Optional[Sequence[PackageMatch]]): List of patterns to match and convert packages.
            packages (Optional[Sequence[PyPackage]]): List of PyPackage objects to create PackageMatch from.
            dest_file (Optional[str]): Path to save the modified pyproject.toml. If None, uses in-place modification.
            in_place (bool): If True, modifies the original file. Cannot be used with dest_file.
            use_toml_sort (bool): If True, sorts the resulting TOML file using toml-sort.
            update (bool): If True, forces an update even if the package is already in a remote state
            backup_dir (Optional[PathType]): Directory to save the backup file. If None, uses a temporary directory.

        Returns:
            List[Tuple[str, str]]: A list of tuples containing the original and modified lines.

        Raises:
            ValueError: If both dest_file and in_place are specified, or if neither is specified.
            ValueError: If both match_patterns and packages are None.
        """
        if match_patterns is None and packages is None:
            raise ValueError("Either match_patterns or packages must be provided")

        if packages is not None:
            match_patterns = self._create_match_patterns_from_packages(
                packages=packages, version_to="{package.version}"
            )
        if not match_patterns:
            raise ValueError(
                "No match patterns were created. pyproject.toml would not be modified."
            )

        return self._convert_to(
            match_patterns=match_patterns,
            dest_file=dest_file,
            in_place=in_place,
            use_toml_sort=use_toml_sort,
            convert_to_local=False,
            update=update,
            backup_dir=backup_dir,
            should_change_module=should_change_module,
        )
    def convert_to_local(
        self,
        match_patterns: Optional[Sequence[PackageMatch]] = None,
        packages: Optional[Sequence[PyPackage]] = None,
        dest_file: Optional[str] = None,
        in_place: bool = False,
        use_toml_sort: bool = True,
        update: bool = False,
        backup_dir: Optional[PathType] = None,
        should_change_module: Callable[[str], bool] = lambda _: True,
    ) -> list[tuple[str, str]]:
        """
        Convert package dependencies to local development versions.

        Args:
            match_patterns (Optional[Sequence[PackageMatch]]): List of patterns to match and convert packages.
            packages (Optional[Sequence[PyPackage]]): List of PyPackage objects to create PackageMatch from.
            dest_file (Optional[str]): Path to save the modified pyproject.toml. If None, uses in-place modification.
            in_place (bool): If True, modifies the original file. Cannot be used with dest_file.
            use_toml_sort (bool): If True, sorts the resulting TOML file using toml-sort.

        Returns:
            List[Tuple[str, str]]: A list of tuples containing the original and modified lines.

        Raises:
            ValueError: If both dest_file and in_place are specified, or if neither is specified.
            ValueError: If both match_patterns and packages are None.
        """
        if match_patterns is None and packages is None:
            raise ValueError("Either match_patterns or packages must be provided")

        all_match_patterns = list(match_patterns or [])

        if packages is not None:
            package_match_patterns = self._create_match_patterns_from_packages(
                packages=packages,
                version_to='{develop = true, path = "{package.path.relative}"}',
            )
            all_match_patterns.extend(package_match_patterns)

        if not all_match_patterns:
            raise ValueError(
                "No match patterns were created. pyproject.toml would not be modified."
            )

        return self._convert_to(
            match_patterns=all_match_patterns,
            dest_file=dest_file,
            in_place=in_place,
            use_toml_sort=use_toml_sort,
            convert_to_local=True,
            update=update,
            backup_dir=backup_dir,
            should_change_module=should_change_module,
        )

    def _create_match_patterns_from_packages(
        self,
        packages: Sequence[PyPackage],
        version_to: str,
    ) -> list[PackageMatch]:
        """
        Create PackageMatch objects from PyPackage objects.

        Args:
            packages (Sequence[PyPackage]): List of PyPackage objects.

        Returns:
            List[PackageMatch]: List of PackageMatch objects created from the provided packages.
        """
        match_patterns = []
        for package in packages:
            # Create a PackageMatch object for each package
            # You may need to adjust this based on your specific requirements
            match_pattern = PackageMatch(
                package_regex=f"^{re.escape(package.name)}$",
                version_regex=r"^.*$",
                version_to=version_to,
            )
            match_patterns.append(match_pattern)
        return match_patterns

    def _convert_to(
        self,
        match_patterns: Sequence[PackageMatch],
        dest_file: Optional[str] = None,
        in_place: bool = False,
        use_toml_sort: bool = True,
        convert_to_local: bool = False,
        update: bool = False,
        backup_dir: Optional[PathType] = None,
        should_change_module: Callable[[str], bool] = lambda _: True,
    ) -> list[tuple[str, str]]:
        """
        Convert package dependencies to remote (non-local) versions.

        This method modifies the pyproject.toml file, changing specified package
        dependencies to use remote versions based on the provided match patterns.

        Args:
            match_patterns (List[PackageMatch]): List of patterns to match and convert packages.
            dest_file (Optional[str]): Path to save the modified pyproject.toml. If None, uses in-place modification.
            in_place (bool): If True, modifies the original file. Cannot be used with dest_file.
            use_toml_sort (bool): If True, sorts the resulting TOML file using toml-sort.

        Returns:
            List[Tuple[str, str]]: A list of tuples containing the original and modified lines.

        Raises:
            ValueError: If both dest_file and in_place are specified, or if neither is specified.
        """
        pyproj = self.pyproj
        assert pyproj is not None
        backup_file = None
        if backup_dir:
            backup_file = os.path.join(backup_dir, f"{pyproj.name}_pyproject.toml")
        if dest_file is not None and in_place:
            raise ValueError("Only one of dest_file or in_place can be specified")
        elif in_place:
            dest_file = str(pyproj.path)
        # elif not in_place and not dest_file:
        #     raise ValueError("destination file is required when in_place is False")

        changes: list[tuple[str, str]] = []
        new_lines: list[str] = []

        with open(pyproj.path) as fp:
            for line in fp:
                if "=" not in line:
                    new_lines.append(line)
                    continue
                sline = line.strip()
                package, previous_package_info = sline.split("=", maxsplit=1)
                package = package.strip()
                previous_package_info = previous_package_info.strip()

                # Check if this module should be changed
                if not should_change_module(package):
                    new_lines.append(line)
                    continue

                is_currently_local = re.search(r"path\s+=", previous_package_info) or re.search(
                    r"develop\s*=\s*true", previous_package_info
                )

                # Skip if we're not forcing an update and the package is already in the desired state
                if not update and (
                    (convert_to_local and is_currently_local)
                    or (not convert_to_local and not is_currently_local)
                ):
                    new_lines.append(line)
                    continue

                new_value = ""
                matched = False

                for mp in match_patterns:
                    m = re.match(mp.package_regex, package)
                    if not m:
                        continue
                    package_name = m.group(0)
                    matched = True
                    try:
                        matched_package = self.packages[package_name]
                    except KeyError:
                        matched_package = None

                    new_value = re.sub(mp.version_regex, mp.version_to, previous_package_info)
                    new_value = special_substitutions(
                        new_value, pkg=pyproj, other_pkg=matched_package
                    )

                if matched:
                    new_line = f"{package} = {new_value}\n"
                    if line != new_line:
                        changes.append((line, new_line))
                        line = new_line

                new_lines.append(line)

        if not changes:
            if use_toml_sort and dest_file:
                run_bash_command(f"toml-sort {dest_file}")
            return changes

        if backup_file:
            os.makedirs(os.path.dirname(backup_file), exist_ok=True)
            shutil.copy(pyproj.path, backup_file)
        if dest_file:
            with tempfile.NamedTemporaryFile("w", delete=False) as tmpfile:
                tmpfile.writelines(new_lines)
                tmpfile.close()
                assert dest_file is not None
                shutil.copy(tmpfile.name, dest_file)
                pyproj._rmlock()
                try:
                    env_helper = EnvHelper(package=pyproj)
                    env_helper.poetry_update()
                    env_helper.poetry_install()
                    # if not convert_to_local:
                    #     # env_helper.poetry_build()
                    #     versions = get_package_versions(pyproj.name)
                    #     ## If on pypi, update the version
                    #     if versions:
                    #         if versions[-1] < pyproj.version:
                    #             env_helper.poetry_publish()
                    #     else:
                    #         # git release
                    #         gm = GithubManager()
                    #         gm.create_github_release_with_dist(pyproj)

                except Exception as e:
                    log.error(f"Poetry update failed: {e}")
                    log.exception(e)
                    raise

        if use_toml_sort and dest_file:
            run_bash_command(f"toml-sort {dest_file}")
        return changes



import logging
import os
import subprocess
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import Any, List, Optional, Tuple

from pi_haiku.models import PathType, PyPackage
from pi_haiku.utils.environment_detector import (
    EnvironmentDetectionError,
    EnvironmentDetector,
    EnvironmentResult,
)
from pi_haiku.utils.utils import run_bash_command

# Define the TRACE level; any integer lower than 10 (DEBUG's level) will work.
TRACE_LEVEL = 5

class CustomLogger(logging.getLoggerClass()):
    def trace(self, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(TRACE_LEVEL):
            self._log(TRACE_LEVEL, message, args, **kwargs)

# Set the custom logger class
logging.setLoggerClass(CustomLogger)

log: CustomLogger = logging.getLogger(__name__)

VERBOSE_LEVEL = 1
class EnvType(Enum):
    VENV = auto()
    CONDA = auto()


@dataclass
class EnvHelper:
    package: PyPackage
    venv_path: Optional[Path] = None
    conda_base_path: Optional[Path] = field(default_factory=lambda: Path.home() / "miniforge3")
    error_file: str = field(init=False)

    def __post_init__(self):
        if isinstance(self.package, str):
            self.package = PyPackage.from_path(self.package)
        self.error_file = f"{self.package.name}_install.log"
        if self.venv_path:
            self.venv_path = Path(self.venv_path)
        if self.conda_base_path:
            self.conda_base_path = Path(self.conda_base_path)

    def has_conda(self) -> bool:
        # Check if the conda environment already exists
        try:
            detect = EnvironmentDetector(self.package, self.venv_path, self.conda_base_path)
            detect_result = detect._detect_conda()
            return detect_result is not None and detect_result.env_type == EnvType.CONDA
        except EnvironmentDetectionError:
            pass
        return False

    def create_conda_project(self) -> bool:
        """
        Create a conda project if it doesn't exist.

        Returns:
            bool: True if the project was created or already exists, False otherwise.
        """
        if self.has_conda():
            log.debug(f"Conda environment already exists for {self.package.name}")
            return True
        # Create the conda environment
        create_command = f"conda create -n {self.package.name} python=3.11 -y"
        return run_bash_command(create_command) == True

    def poetry_update(self) -> Optional[str]:
        try:
            detect = EnvironmentDetector(self.package, self.venv_path, self.conda_base_path)
            env_result = detect.detect_environment()
            command = f"{env_result.activate_command} && poetry update -vvv"
            sh_result = run_bash_command(command, cwd=self.package.path.parent)
            if command:
                if "No dependencies to install or update" in sh_result.stdout:
                    log.debug(f"No dependencies to install or update for {self.package.name}")
                    return None
                log.trace(sh_result.stdout)
                log.debug(
                    f"Update successful for {self.package.name} v{self.package.version} using {env_result.env_type} environment"
                )
            return sh_result.stdout
        except EnvironmentError as e:
            log.error(f"Installation failed for {self.package.name}: {e}")
        except subprocess.CalledProcessError as e:
            log.error(
                f"Installation command failed for {self.package.name}. Check {self.error_file} for details."
            )
        return None

    def poetry_install(self) -> Optional[str]:
        try:
            detect = EnvironmentDetector(self.package, self.venv_path, self.conda_base_path)
            env_result = detect.detect_environment()
            command = f"{env_result.activate_command} && poetry install -vvv"
            sh_result = run_bash_command(command, cwd=self.package.path.parent)
            if command:
                log.trace(sh_result.stdout)
                log.debug(
                    f"Install successful for {self.package.name} v{self.package.version} using {env_result.env_type} environment"
                )
            return sh_result.stdout
        except EnvironmentError as e:
            log.error(f"Installation failed for {self.package.name}: {e}")
        except subprocess.CalledProcessError as e:
            log.error(
                f"Installation command failed for {self.package.name}. Check {self.error_file} for details."
            )
        return None

    @staticmethod
    def from_path(path: PathType) -> "EnvHelper":
        return EnvHelper(PyPackage.from_path(path))



import os
from pathlib import Path
from typing import Optional

from git import Repo
from github import Github
from pi_conf import ConfigSettings
from pydantic import SecretStr, BaseModel, Field


from pi_haiku import PyPackage


class Indexes(BaseModel):
    uri: str
    access_token: SecretStr


class GitHubSettings(ConfigSettings):
    indexes: list[Indexes] = Field(default_factory=list, alias="index")

    model_config = {
        "appname": "pi-haiku",
    }


class GithubManager:
    def __init__(self, settings: Optional[GitHubSettings] = None):
        self.settings = settings or GitHubSettings()
        self.access_map = {index.uri: index.access_token for index in self.settings.indexes}

    @staticmethod
    def get_github_url(package: PyPackage) -> Optional[str]:
        try:
            repo = Repo(package.path.parent, search_parent_directories=False)
            remote_url = repo.remotes.origin.url

            # Remove .git extension if present
            github_url = remote_url.rstrip(".git")
            return github_url
        except Exception as e:
            print(f"Error: {e}")
            return None

    def create_github_release_with_dist(self, package: PyPackage, release_body: str = ""):
        repo_url = self.get_github_url(package)
        if not repo_url:
            raise ValueError("Could not determine the GitHub repository name")

        # Extract the owner url
        owner_url = repo_url.rsplit("/", 1)[0]

        # Extract the owner/repo part
        repo_name = "/".join(repo_url.split("/")[-2:])

        access_token = self.access_map[owner_url].get_secret_value()
        tag_name = f"v{package.version}"
        release_name = package.version
        g = Github(access_token)
        repo = g.get_repo(repo_name)

        # Create the release
        release = repo.create_git_release(
            tag=tag_name, name=release_name, message=release_body, draft=False, prerelease=False
        )

        return release


# pkg = PyPackage.from_path("~/workspace/qrev-cache")
# gm = GithubManager()
# gm.create_github_release_with_dist(pkg)


# print(pkg, pkg.path)
# print(GithubManager.get_github_repo_name(pkg))
# # Example usage
# token = "your_github_personal_access_token"
# repo_name = "owner/repo"
# tag_name = "v1.0.0"
# release_name = "Version 1.0.0"
# release_body = "Description of this release"
# dist_path = "dist"  # This is where Poetry typically puts build artifacts

# # Assuming you've already run `poetry build`
# try:
#     release = create_github_release_with_dist(token, repo_name, tag_name, release_name, release_body, dist_path)
#     print(f"Release created successfully: {release.html_url}")
#     print("Distribution files uploaded.")
# except Exception as e:
#     print(f"An error occurred: {str(e)}")



import json
import os
import subprocess
import sys
from dataclasses import dataclass
from logging import getLogger
from subprocess import CalledProcessError
from typing import Any, Optional, Sequence

from pi_haiku.models import PathType, PyPackage

log = getLogger(__name__)


@dataclass
class CommandResult:
    success: bool
    stdout: str
    stderr: str

    def __bool__(self):
        return self.success


def get_conda_info() -> dict[str, Any]:
    try:
        result = subprocess.run(
            ["conda", "info", "--json"], check=True, capture_output=True, text=True
        )
        return json.loads(result.stdout)
    except subprocess.CalledProcessError as e:
        log.error(f"Error getting conda info: {e}")
        log.error(f"Error output: {e.stderr}")
        raise


def run_bash_command(
    command: str,
    output_file: Optional[str] = None,
    error_file: Optional[str] = None,
    cwd: Optional[PathType] = None,
    verbose: bool = True,
    check: bool = True,
) -> CommandResult:
    if cwd:
        cwd = str(cwd)
    try:
        if verbose:
            print(f"Cmd: {command}, cwd={cwd}", flush=True)
        conda_info = get_conda_info()
        activate_path = os.path.join(conda_info["root_prefix"], "bin", "activate")
        command = f"source {activate_path} && {command}"
        result = subprocess.run(
            command,
            shell=True,
            check=check,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            cwd=cwd,
        )

        if result.stdout and output_file:
            with open(output_file, "w") as f:
                f.write(result.stdout)
            if verbose:
                print(f"Command Output: {result.stdout}")
        if result.stderr and error_file:
            with open(error_file, "w") as f:
                f.write(result.stderr)
            if verbose:
                print("Command Error Output:\n", result.stderr)

        success = result.returncode == 0
        return CommandResult(success, result.stdout, result.stderr)

    except CalledProcessError as e:
        return CommandResult(False, e.stdout, e.stderr)


def _check_value_is_primitive(value: str) -> None:
    primitives = (bool, str, int, float, type(None))
    try:
        js = json.loads(value)
        if not isinstance(js, primitives):
            raise ValueError(
                f"Advanced types, i.e array or dict, are currently not supported. Got {js}"
            )
    except json.JSONDecodeError:
        print(f"Could not parse the package info: {value}", file=sys.stderr)
        raise


def custom_sort_dict(input_dict: dict[str, Any], order_list: list[str]) -> dict[str, Any]:
    def sort_key(key: str) -> tuple[int, int | str]:
        if key in order_list:
            return (0, order_list.index(key))
        else:
            return (1, key)

    sorted_keys = sorted(input_dict.keys(), key=sort_key)
    sorted_dict = {key: input_dict[key] for key in sorted_keys}
    return sorted_dict


def create_dag(packages: Sequence[PyPackage]) -> dict[str, set[str]]:
    dag: dict[str, set[str]] = {}
    for package in packages:
        if package.name not in dag:
            dag[package.name] = set()
        for dep in package.dependencies.keys():
            if dep not in dag:
                dag[dep] = set()
            dag[dep].add(package.name)
    return dag


def topological_sort(dag: dict[str, set[str]]) -> list[str]:
    def visit(node: str) -> None:
        if node in temp_mark:
            raise ValueError(f"Cycle detected: {node}")
        if node not in perm_mark:
            temp_mark.add(node)
            for neighbor in dag.get(node, set()):
                visit(neighbor)
            temp_mark.remove(node)
            perm_mark.add(node)
            result.append(node)

    temp_mark: set[str] = set()
    perm_mark: set[str] = set()
    result: list[str] = []

    for node in dag:
        if node not in perm_mark:
            visit(node)

    return result[::-1]  # reverse the result to get the correct order


def special_substitutions(s: str, pkg: PyPackage, other_pkg: Optional[PyPackage]) -> str:
    if "{package}" in s:
        s = s.replace("{package}", pkg.name)
    if "{version}" in s:
        s = s.replace("{version}", f'"{pkg.version}"')
    if "{package." in s and other_pkg is None:
        raise ValueError("other_pkg is required for special substitutions requiring other package")
    assert other_pkg  # to make mypy happy as it is already checked in the above line
    if "{package.version}" in s:
        caret_str = "^" if not "^" in other_pkg.version else ""
        s = s.replace("{package.version}", f'"{caret_str}{other_pkg.version}"')
    if "{package.path.relative}" in s:
        s = s.replace("{package.path.relative}", str(other_pkg.relative_to_package(pkg).parent))
    if "{package.path.absolute}" in s:
        s = s.replace("{package.path.absolute}", str(other_pkg.path.parent))
    ## use a regex to remove common mistakes such as double caret
    s = s.replace("^+", "^")
    return s


def find_duplicates(lst: Sequence[Any]) -> list[Any]:
    seen = set()
    duplicates = set()

    for item in lst:
        if item in seen:
            duplicates.add(item)
        else:
            seen.add(item)

    return list(duplicates)



import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
import tomllib as toml
from dataclasses import dataclass, field
from enum import StrEnum
from pathlib import Path
from typing import Any, Optional, Self, TypeVar

PathType = str | Path

MATCH_ALL: str = r"^.*$"


class BuildSystemError(Exception):
    pass


@dataclass
class PackageMatch:
    """
    A class to match a package and change the version
    Special variables:
        {package} - The source package name
        {version} - The source version
        {package.version} - The matching package version
        {package.path.relative} - The relative path to the package from src
        {package.path.absolute} - The absolute path to the package
    """

    package_regex: str
    version_regex: str = MATCH_ALL
    version_to: str = "{package.version}"


@dataclass
class ToLocalMatch(PackageMatch):
    version_to: str = '{develop = true, path = "{package.path.relative}"}'


@dataclass
class ToRemoteMatch(PackageMatch):
    version_to: str = "{package.version}"


class BuildSystem(StrEnum):
    SETUPTOOLS = "setuptools"
    POETRY = "poetry"
    FLIT = "flit"
    HATCHLING = "hatchling"
    PDM = "pdm"


@dataclass
class PyPackage:
    name: str
    version: str
    path: Path
    dependencies: dict[str, Any] = field(default_factory=dict)
    toml_data: dict[str, Any] = field(default_factory=dict)

    @staticmethod
    def get_dependencies(file_path: PathType) -> dict[str, Any]:
        with open(file_path, "rb") as file:
            data: dict[str, Any] = toml.load(file)

        dependencies: dict[str, Any] = {}
        if _get_build_system(data, BuildSystem.POETRY, None):
            if "dependencies" in data["tool"]["poetry"]:
                dependencies.update(data["tool"]["poetry"]["dependencies"])

            if "group" in data["tool"]["poetry"]:
                for group, group_data in data["tool"]["poetry"]["group"].items():
                    if "dependencies" in group_data:
                        dependencies.update(group_data["dependencies"])
        elif _get_build_system(data, BuildSystem.SETUPTOOLS, None) and "project" in data:
            if "dependencies" in data["project"]:
                dependencies.update({dep: None for dep in data["project"]["dependencies"]})

        return dependencies
    
    @staticmethod
    def from_path(path: PathType) -> "PyPackage":
        path = Path(path).expanduser().resolve()
        if path.is_dir():
            path = path / "pyproject.toml"
        with open(path, "rb") as fp:
            data: dict[str, Any] = toml.load(fp)
        dependencies = PyPackage.get_dependencies(path)
        build = _get_build_system(data, BuildSystem.POETRY, None)
        if build == BuildSystem.POETRY:
            poetry = data["tool"]["poetry"]
            return PyPackage(
                name=poetry["name"],
                version=poetry["version"],
                path=path,
                dependencies=dependencies,
                toml_data=data,
            )
        else:
            try:
                proj = data["project"]
                return PyPackage(
                    name=proj["name"],
                    version=proj["version"],
                    path=path,
                    dependencies=dependencies,
                    toml_data=data,
                )
            except:
                print(f"Could not read the {path} file", file=sys.stderr)
                raise

    def get_local_dependencies(self) -> dict[str, str]:
        local_deps: dict[str, str] = {}
        for dep, version in self.dependencies.items():
            if isinstance(version, dict) and "path" in version:
                local_deps[dep] = version["path"]
        return local_deps

    def __str__(self) -> str:
        return f"{self.name}=={self.version}"

    def __repr__(self) -> str:
        return str(self)

    def relative_to_package(self, other: Self) -> Path:
        return Path(os.path.relpath(self.path, other.path.parent))

    def _rmlock(self, missing_ok: bool = True) -> None:
        lockfile = self.path.parent / "poetry.lock"
        if lockfile.exists():
            lockfile.unlink(missing_ok=missing_ok)

    def __hash__(self) -> int:
        return hash(self.name)


def _get_build_system(
    data: dict[str, Any], search_for: Optional[BuildSystem] = None, default: Optional[Any] = None
) -> BuildSystem:
    build_system = data.get("build-system", {})
    for build_system in data["build-system"]["requires"]:
        v: Optional[BuildSystem] = None
        if "poetry-core" in build_system:
            v = BuildSystem.POETRY
        elif "setuptools" in build_system:
            v = BuildSystem.SETUPTOOLS
        elif "flit" in build_system:
            v = BuildSystem.FLIT
        elif "hatchling" in build_system:
            v = BuildSystem.HATCHLING
        elif "pdm" in build_system:
            v = BuildSystem.PDM
        if not search_for and v:
            return v
        if search_for and v is not None and v == search_for:
            return v
    else:
        if search_for:
            raise BuildSystemError(f"Could not find the build system: {search_for}")

    if default:
        return default
    raise BuildSystemError(f"Unsupported build system: {build_system}")



import logging
import os
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import Any, List, Optional, Tuple

from pi_haiku.models import PyPackage
from pi_haiku.utils.utils import run_bash_command

log = logging.getLogger(__name__)

class EnvType(Enum):
    VENV = auto()
    CONDA = auto()



@dataclass
class EnvironmentResult:
    env_type: EnvType
    activate_command: str


class EnvironmentDetectionError(EnvironmentError):
    pass


class EnvironmentDetector:
    def __init__(
        self,
        package: Optional[PyPackage] = None,
        venv_path: Optional[Path] = None,
        conda_base_path: Optional[Path] = None,
        package_path: Optional[Path] = None,
    ):
        if package_path:
            package = PyPackage.from_path(package_path)
        self.package = package
        if self.package and not isinstance(self.package, PyPackage):
            self.package = PyPackage.from_path(self.package)
        if not self.package:
            raise ValueError(f"No package provided. {package}, {package_path}")
        if not isinstance(self.package, PyPackage):
            raise ValueError(f"Invalid package provided. {package}, {package_path}")
        self.venv_path = venv_path
        self.conda_base_path = conda_base_path
        

    def detect_environment(self) -> EnvironmentResult:
        venv_result = self._detect_venv()
        if venv_result:
            return venv_result

        conda_result = self._detect_conda()
        if conda_result:
            return conda_result

        log.error("No valid virtual environment found.")
        raise EnvironmentDetectionError("No valid virtual environment found.")

    def _detect_venv(self) -> Optional[EnvironmentResult]:
        venv_names = [".venv", "venv", "env"]
        venv_locations = [self.venv_path] if self.venv_path else []
        assert self.package
        venv_locations.extend([self.package.path.parent / name for name in venv_names])
        venv_locations.extend([self.package.path.parent.parent / name for name in venv_names])

        for venv_loc in venv_locations:
            if venv_loc and venv_loc.is_dir() and self._is_valid_environment(venv_loc):
                activate_path = self._get_activate_path(venv_loc)
                if activate_path:
                    log.info(f"Found venv at {venv_loc}")
                    return EnvironmentResult(EnvType.VENV, str(activate_path))
        return None

    def _detect_conda(self, include_base: bool = False) -> Optional[EnvironmentResult]:
        if not self.conda_base_path:
            return None
        assert self.package

        result = run_bash_command(f"conda env list | grep {self.package.name}")
        env_list = result.stdout.strip().split('\n')
        # Parse the output to get environment names and paths
        conda_envs = {}
        for line in env_list:  # Skip the first two lines (header)
            if "#" in line:
                continue
            if line.strip():
                parts = line.split()
                env_name = parts[0]
                env_path = parts[-1]
                conda_envs[env_name] = env_path

        env_names = [self.package.name]
        if include_base:
            env_names.append("base")

        for env_name in env_names:
            if env_name in conda_envs:
                log.info(f"Found conda environment: {env_name}")
                return EnvironmentResult(EnvType.CONDA, f"conda activate {env_name}")
            conda_env_path = self.conda_base_path / "envs" / env_name
            if conda_env_path.is_dir() and self._is_valid_environment(conda_env_path):
                log.info(f"Found conda environment: {env_name}")
                return EnvironmentResult(EnvType.CONDA, f"conda activate {env_name}")

        # If no environment found in 'envs', check if conda_base_path itself is a valid environment
        if self._is_valid_environment(self.conda_base_path):
            log.info("Found conda base environment")
            return EnvironmentResult(EnvType.CONDA, "conda activate base")

        return None

    @staticmethod
    def _get_activate_path(env_path: Path) -> Optional[Path]:
        if os.name == "nt":  # Windows
            activate_path = env_path / "Scripts" / "activate"
        else:  # Unix-like systems
            activate_path = env_path / "bin" / "activate"

        return activate_path if activate_path.exists() else None

    def _is_valid_environment(self, env_path: Path) -> bool:
        # Simplified check for testing purposes
        return (env_path / "bin" / "activate").exists() or (
            env_path / "Scripts" / "activate"
        ).exists()


from pi_haiku.pyproject_modifier import PyProjectModifier
from pi_haiku.models import PackageMatch, PyPackage, ToLocalMatch, ToRemoteMatch

__all__ = [
    "PackageMatch",
    "PyProjectModifier",
    "PyPackage",
    "ToLocalMatch",
    "ToRemoteMatch",
]



import os
import tempfile
from typing import Callable, Optional, Union

from pi_haiku import (
    PackageMatch,
    PyPackage,
    PyProjectModifier,
    ToLocalMatch,
    ToRemoteMatch,
)
from pi_haiku.models import PathType
from pi_haiku.utils.environment_utils import EnvHelper
from pi_haiku.utils.utils import (
    create_dag,
    custom_sort_dict,
    run_bash_command,
    topological_sort,
)

PackageType = Union[PyPackage, str]


class Haiku:
    """
    A class for managing and converting Python projects between local and remote configurations.
    """

    @staticmethod
    def install(proj_path: Optional[PathType] = None, pkg: Optional[PyPackage] = None) -> None:
        """
        Install a package in the specified project.
        """
        if proj_path is None and pkg is None:
            raise ValueError("Either proj_path or pkg must be specified")

        if proj_path is not None:
            pkg = PyPackage.from_path(proj_path)
        if pkg is None:
            raise ValueError(f"Package could not be found from proj_path: {proj_path}, pkg: {pkg}")
        main_pkg = pkg
        local_deps = main_pkg.get_local_dependencies()
        all_local_deps: dict[str, PyPackage] = {}
        for dep in local_deps:
            projs = PyProjectModifier.find_pyprojects(dep)
            all_local_deps.update(projs)
        dag = create_dag(list(all_local_deps.values()))
        flattened = topological_sort(dag)
        flattened = [p for p in flattened if p in all_local_deps]
        local_projs_to_install = [all_local_deps[p] for p in flattened]
        for proj in local_projs_to_install:
            eh = EnvHelper(proj)
            eh.poetry_install()
        eh = EnvHelper(main_pkg)
        eh.poetry_install()

    @staticmethod
    def _convert_projects(
        dir: PathType,
        convert_function: Callable,
        exclude_projects: Optional[list[PackageType]] = None,
        include_projects: Optional[list[PackageType]] = None,
        only_change_projects: Optional[list[PackageType]] = None,
        source_projects: Optional[list[PackageType]] = None,
        dry_run: bool = True,
        verbose: bool = False,
        update: bool = False,
        backup_dir: Optional[PathType] = None,
    ) -> dict[PyPackage, list[tuple[str, str]]]:
        """
        Internal method to convert projects using the specified conversion function.
        """
        projs = PyProjectModifier.find_pyprojects(dir)
        changes: dict[PyPackage, list[tuple[str, str]]] = {}
        dag = create_dag(list(projs.values()))

        flattened = topological_sort(dag)
        flattened = [p for p in flattened if p in projs]

        list_projs = list(projs.values())
        should_print = verbose or dry_run

        def get_project_name(project: PackageType) -> str:
            return project.name if isinstance(project, PyPackage) else project

        def should_process_project(proj_name: str) -> bool:
            if exclude_projects and any(get_project_name(p) == proj_name for p in exclude_projects):
                return False
            if include_projects and not any(
                get_project_name(p) == proj_name for p in include_projects
            ):
                return False
            return True

        def should_change_module(module_name: str) -> bool:
            if only_change_projects:
                v = any(get_project_name(p) == module_name for p in only_change_projects)
                return v
            return True

        for proj_name in flattened:
            if not should_process_project(proj_name):
                # if should_print:
                #     print(f"Skipped project {proj_name} due to include/exclude restrictions")
                continue

            proj = projs[proj_name]
            if should_print:
                print(f"        =============== {proj} =============== ")

            pmod = PyProjectModifier(proj.path, packages=projs)

            # Apply the conversion function to the project
            file_changes = convert_function(
                pmod,
                packages=list_projs,
                use_toml_sort=False,
                update=update,
                in_place=not dry_run,
                backup_dir=backup_dir if not dry_run else None,
                should_change_module=should_change_module,  # Pass this function to control which modules are changed
            )

            changes[proj] = file_changes

            if should_print and file_changes:
                for c in file_changes:
                    from_str, to_str = c[0].strip(), c[1].strip()
                    print(f"{from_str}  ->  {to_str}")

        return changes

    @staticmethod
    def convert_projects_to_local(
        dir: PathType,
        exclude_projects: Optional[list[PackageType]] = None,
        include_projects: Optional[list[PackageType]] = None,
        only_change_projects: Optional[list[PackageType]] = None,
        dry_run: bool = True,
        verbose: bool = False,
        backup_dir: Optional[PathType] = None,
    ) -> dict[PyPackage, list[tuple[str, str]]]:
        """
        Convert multiple projects to local configuration.
        """
        return Haiku._convert_projects(
            dir=dir,
            convert_function=PyProjectModifier.convert_to_local,
            exclude_projects=exclude_projects,
            include_projects=include_projects,
            only_change_projects=only_change_projects,
            dry_run=dry_run,
            verbose=verbose,
            backup_dir=backup_dir,
        )

    @staticmethod
    def convert_projects_to_remote(
        dir: PathType,
        exclude_projects: Optional[list[PackageType]] = None,
        include_projects: Optional[list[PackageType]] = None,
        only_change_projects: Optional[list[PackageType]] = None,
        dry_run: bool = True,
        verbose: bool = False,
        update: bool = False,
        backup_dir: Optional[PathType] = None,
    ) -> dict[PyPackage, list[tuple[str, str]]]:
        """
        Convert multiple projects to remote configuration.
        """
        return Haiku._convert_projects(
            dir=dir,
            convert_function=PyProjectModifier.convert_to_remote,
            exclude_projects=exclude_projects,
            include_projects=include_projects,
            only_change_projects=only_change_projects,
            dry_run=dry_run,
            verbose=verbose,
            update=update,
            backup_dir=backup_dir,
        )



import json

import requests


class PackageNotFoundError(Exception):
    pass


def get_package_versions(package_name) -> list[str] | None:
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        response = requests.get(url, timeout=5)
        if response.status_code == 200:
            data = response.json()
            versions = list(data["releases"].keys())
            versions.sort(key=lambda s: [int(u) for u in s.split(".")])
            return versions
        elif response.status_code == 404:
            return None
        else:
            print(f"Unexpected response code: {response.status_code}")
    except requests.exceptions.RequestException as e:
        print(f"Error connecting to PyPI: {e}")
    except json.JSONDecodeError:
        print("Error decoding JSON response")
    return None
